/*
 * dupfind - find duplicate files
 *
 * Copyright 2005, 2006  Yoshizawa Takashi  <morca@yahoo.co.jp>
 * http://www.geocities.jp/morca/
 * Last hack: June 19, 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License (Version 2,
 *  June 1991) as published by the Free Software Foundation.  At the
 *  time of writing, that license was published by the FSF with the URL
 *  http://www.gnu.org/copyleft/gpl.html, and is incorporated herein by
 *  reference.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#ifdef _WIN32
#include <dir.h>
#include <wildargs.h>
#else
#include <unistd.h>
#endif

#define BLOCK_HASH

#ifdef _WIN32
#define PATH_DELIM	'\\'
#else
#define PATH_DELIM	'/'
#endif

struct node {
	char *name;
	mode_t mode;
	off_t size;
	dev_t dev;
	ino_t ino;
	int disp;
#define HASH_AVAIL	0x80000000
#define HASH_ERROR	0x40000000
	unsigned long hash;
#ifdef BLOCK_HASH
	unsigned long *block;
#endif
	struct node *parent;
	struct node *next;
	struct node *prev;
} *Head, *Last;
char *Arg0;
int Files;

int end_with_slash(char *path)
{
#ifdef _WIN32
#define is1kanji(c)	(0x81 <= (c) && ((c) <= 0x9f || 0xe0 <= (c) && (c) <= 0xef))
#define is2kanji(c)	(0x40 <= (c) && (c) <= 0xfc && (c) != 0x7f)
	unsigned char *p, *slash;

	/*
	 * C:\	don't append backslash
	 * C:	don't append backslash
	 * dir	append backslash
	 * dir\	don't append backslash
	 */
	for (p = (unsigned char *)path, slash = NULL; *p; p++) {
		if (is1kanji(*p) && is2kanji(*(p + 1)))	/* Shift_JIS code */
			p++;
		else if (*p == PATH_DELIM || *p == '/' || *p == ':')
			slash = p;
	}
	return p == slash + 1;
#else
	return path[strlen(path) - 1] == PATH_DELIM;
#endif
}

static void setpath(char *path, char *name, struct node *parent)
{
	struct node *node;

	strcpy(path, name);
	for (node = parent; node; node = node->parent) {
//@@ len check
		if (!end_with_slash(node->name)) {
			memmove(path + 1, path, strlen(path) + 1);
			path[0] = PATH_DELIM;
		}
		memmove(path + strlen(node->name), path, strlen(path) + 1);
		memcpy(path, node->name, strlen(node->name));
	}
}

static int entlist(char *name, struct node *parent, int disp)
{
#ifndef MAXPATH
#define MAXPATH	8192
#endif
	static char path[MAXPATH] = "";
	int len = strlen(path);
	struct stat buf;
	struct node *node;

	if (len > 0 && !end_with_slash(path)) {
		path[len] = PATH_DELIM;
		path[len + 1] = '\0';
	}
	strcat(path, name);
#ifdef _WIN32
	if (strlen(name) == 2 && name[1] == ':') {
		strcat(path, ".");
		name = path + strlen(path) - 3;
	}
#endif
#ifdef _WIN32
	if (stat(name, &buf) < 0)
//@@ dir\ NG
#else
	if (lstat(name, &buf) < 0)
#endif
	{
		perror(path);
		return -1;
	}
	if ((node = malloc(sizeof(struct node))) == NULL) {
		perror(Arg0);
		exit(1);
	}
	if ((node->name = strdup(name)) == NULL) {
		perror(Arg0);
		exit(1);
	}
	node->mode = buf.st_mode;
	node->size = buf.st_size;
	node->dev = buf.st_dev;
	node->ino = buf.st_ino;
	node->disp = disp;
	node->parent = parent;
	node->hash = 0;
#ifdef BLOCK_HASH
	node->block = NULL;
#endif
	node->next = NULL;
	node->prev = NULL;
	if (Head == NULL)
		Head = node;
	if (Last != NULL)
		Last->next = node;
	Last = node;
//	printf("%s\n", path);
	if (!S_ISDIR(buf.st_mode)) {
		path[len] = '\0';
		return 0;
	}
	if (!chdir(name)) {
		DIR *dir;

//TODO: \\xxx\xxx
		if ((dir = opendir(".")) != NULL) {
			struct dirent *dirent;
			struct temp {
				char *name;
				struct temp *next;
			} *head = NULL, *p;

			while ((dirent = readdir(dir)) != NULL) {
				if (!strcmp(dirent->d_name, ".")
					|| !strcmp(dirent->d_name, ".."))
					continue;
				if ((p = malloc(sizeof(struct temp))) == NULL) {
					perror(Arg0);
					continue;
				}
				if ((p->name = strdup(dirent->d_name)) == NULL) {
					perror(Arg0);
					free(p);
					continue;
				}
				p->next = head;
				head = p;
			}
			closedir(dir);
			for (p = head; p; head = p, p = p->next, free(head)) {
				entlist(p->name, node, disp);
				free(p->name);
			}
		} else
			perror(path);
		if (chdir(".."))
			perror("..");
	} else
		perror(path);
	path[len] = '\0';
	return 0;
}

#define CRCPOLY2	0x8408U
static unsigned short crctable[UCHAR_MAX + 1];

static void initcrc(void)
{
	unsigned short i, j, r;

	for (i = 0; i <= UCHAR_MAX; i++) {
		r = i;
		for (j = 0; j < CHAR_BIT; j++)
			if (r & 1)
				r = (r >> 1) ^ CRCPOLY2;
			else
				r >>= 1;
		crctable[i] = r;
	}
}

static unsigned short crc16(int n, unsigned char *c, unsigned short last)
{
	unsigned short r;

	r = last ^ 0xffffU;
	while (0 <= --n)
		r = (r >> CHAR_BIT) ^ crctable[(unsigned char)r ^ *c++];
	return r ^ 0xffffU;
}

static int comp_file(const void *arg1, const void *arg2)
{
	struct node *node1 = *(struct node **)arg1;
	struct node *node2 = *(struct node **)arg2;
#define BUFSIZE	8192
	unsigned char buf1[BUFSIZE], buf2[BUFSIZE];
	char path1[MAXPATH], path2[MAXPATH];
	FILE *fp1, *fp2;
	unsigned short crc1, crc2;
	off_t total;
	size_t nread;
	int ret;
#ifdef BLOCK_HASH
	int i;
#endif

	if (node1->size != node2->size)
		return node1->size - node2->size;
//#define NOMATCH	comp_disp_path(&node1, &node2)
#define NOMATCH 1
	if ((node1->hash & HASH_ERROR) || (node2->hash & HASH_ERROR))
		return NOMATCH;
	if ((node1->hash & HASH_AVAIL) && (node2->hash & HASH_AVAIL)
		&& node1->hash != node2->hash)
		return node1->hash - node2->hash;
	if (node2->next != NULL || node2->prev != NULL) {
		struct node *n;

		for (n = node2; n->prev != NULL; n = n->prev)
			;
		for (; n != NULL; n = n->next)
			if (node1 == n)
				return 0;	/* already in the list */
	}
#ifdef BLOCK_HASH
	for (i = 0; i * BUFSIZE < node1->size; i++) {
		if (!(node1->block[i] & HASH_AVAIL) || !(node2->block[i] & HASH_AVAIL))
			break;
		if (node1->block[i] != node2->block[i])
			return node1->block[i] - node2->block[i];
	}
#endif
	setpath(path1, node1->name, node1->parent);
	if ((fp1 = fopen(path1, "rb")) == NULL) {
		perror(path1);
		node1->hash |= HASH_ERROR;
		return NOMATCH;
	}
	setpath(path2, node2->name, node2->parent);
	if ((fp2 = fopen(path2, "rb")) == NULL) {
		perror(path2);
		node2->hash |= HASH_ERROR;
		fclose(fp1);
		return NOMATCH;
	}
	crc1 = 0, crc2 = 0;
	ret = 0;
	total = 0;
#ifndef BLOCK_HASH
	while (!feof(fp1)) {
#else
	for (i = 0; !feof(fp1); i++) {
#endif
		nread = fread(buf1, 1, sizeof(buf1), fp1);
		if (ferror(fp1)) {
			perror(path1);
			node1->hash |= HASH_ERROR;
			ret = NOMATCH;
			break;
		}
		if ((nread < sizeof(buf1) && total + (off_t)nread < node1->size)
			|| total + (off_t)nread > node1->size) {
			fprintf(stderr, "%s: Volatile\n", path1);
			node1->hash |= HASH_ERROR;
			ret = NOMATCH;
			break;
		}
		if (nread == 0)
			break;
#ifdef BLOCK_HASH
		if (!(node1->block[i] & HASH_AVAIL)) {
			crc1 = crc16(nread, buf1, crc1);
			node1->block[i] = crc1 | HASH_AVAIL;
		} else
			crc1 = node1->block[i];
//@@re-calc to detect volatile
#endif
		if (fread(buf2, 1, sizeof(buf2), fp2) < nread) {
			fprintf(stderr, "%s: Volatile\n", path2);
			node2->hash |= HASH_ERROR;
			ret = NOMATCH;
			break;
		}
		if (ferror(fp2)) {
			perror(path2);
			node2->hash |= HASH_ERROR;
			ret = NOMATCH;
			break;
		}
#ifdef BLOCK_HASH
		if (!(node2->block[i] & HASH_AVAIL)) {
			crc2 = crc16(nread, buf2, crc2);
			node2->block[i] = crc2 | HASH_AVAIL;
		} else
			crc2 = node2->block[i];
		if (crc1 != crc2) {
			ret = crc1 - crc2;
			break;
		}
#endif
		total += nread;
		if (total > node1->size) {
			ret = NOMATCH;
			break;
		}
		ret = memcmp(buf1, buf2, nread);
		if (ret)
			break;
#ifndef BLOCK_HASH
		crc1 = crc16(nread, buf1, crc1);
		crc2 = crc16(nread, buf2, crc2);
#endif
	}
	if (ret == 0) {
		struct node *n1, *n2;

		node1->hash = HASH_AVAIL & crc1;
		node2->hash = HASH_AVAIL & crc2;

		/* doubly linked list */
		for (n1 = node1; n1->next; n1 = n1->next)
			;
		for (n2 = node2; n2->next; n2 = n2->next)
			;
		if (n1 != n2) {
			/* different lists, merge them */
			for (n2 = node2; n2->prev; n2 = n2->prev)
				;
			n1->next = n2;
			n2->prev = n1;
		}
	}
	fclose(fp1);
	fclose(fp2);
	return ret;
}

static int comp_size_ino(const void *arg1, const void *arg2)
{
	const struct node *node1 = *(struct node **)arg1;
	const struct node *node2 = *(struct node **)arg2;
#ifdef _WIN32
	char path[MAXPATH], path1[MAXPATH], path2[MAXPATH];
	int cmp;
#endif

	if (node1->size != node2->size)
		return node1->size - node2->size;
#ifdef _WIN32
	setpath(path, node1->name, node1->parent);
	GetFullPathName(path, sizeof(path1), path1, NULL);
	setpath(path, node2->name, node2->parent);
	GetFullPathName(path, sizeof(path2), path2, NULL);
	if ((cmp = stricmp(path1, path2)) != 0)
		return cmp;
#else
	if (node1->dev != node2->dev)
		return node1->dev - node2->dev;
	if (node1->ino != node2->ino)
		return node1->ino - node2->ino;
#endif
	return node1->disp - node2->disp;
}

static int comp_disp_path(const void *arg1, const void *arg2)
{
	const struct node *node1 = *(struct node **)arg1;
	const struct node *node2 = *(struct node **)arg2;
	char path1[MAXPATH], path2[MAXPATH];

	if (node1->disp != node2->disp)
		return node2->disp - node1->disp;
	setpath(path1, node1->name, node1->parent);
	setpath(path2, node2->name, node2->parent);
#ifdef _WIN32
	return stricmp(path1, path2);
#else
	return strcmp(path1, path2);
#endif
}

static void free_node(struct node *node)
{
	if (node == NULL)
		return;
	if (node->name != NULL)
		free(node->name);
#ifdef BLOCK_HASH
	if (node->block != NULL)
		free(node->block);
#endif
	free(node);
}

int main(int argc, char **argv)
{
	int i, j, n;
	struct node *node, *next;
	struct node **list = NULL;
	char path[MAXPATH];
	int disp = 1, all = 0, len = 0, uniq = 0, count = 0, debug = 0;

	Arg0 = argv[0];
	initcrc();

	/* find all entries */
	getcwd(path, sizeof(path));
	for (i = 1; i < argc; i++)
			if (!strcmp(argv[i], "--debug"))
				debug = 1;
	for (i = 1; i < argc; i++)
		if (argv[i][0] == '-') {
			if (!strcmp(argv[i], "-q"))
				disp = 0;
			else if (!strcmp(argv[i], "-a"))
				all = 1;
			else if (!strcmp(argv[i], "-l"))
				len = 1;
			else if (!strcmp(argv[i], "-u"))
				uniq = 1;
			else if (!strcmp(argv[i], "--debug"))
				;
			else {
usage:
				fprintf(stderr, "usage: %s [-a] [-l] [-u] [--debug] path... [-q path...]\n", argv[0]);
				return 1;
			}
		} else if (argv[i][0] == '\0') {
			goto usage;
		} else {
			if (debug) fprintf(stderr, "look path (%s)\n", argv[i]);
			entlist(argv[i], NULL, disp);
			if (chdir(path))
				perror(path);
			if (disp)
				count++;
		}
	if (count == 0) {
		if (debug) fprintf(stderr, "look path (%s)\n", ".");
		entlist(".", NULL, 1);
		if (chdir(path))
			perror(path);
	}

	/* keep regular files */
	for (node = Head, n = 0; node; node = node->next)
		if (S_ISREG(node->mode))
			n++;
	if (debug) fprintf(stderr, "regular files (%d)\n", n);
	if (n == 0)
		goto end;
	if (n && (list = malloc(sizeof(list[0]) * n)) == NULL) {
		perror(Arg0);
		exit(1);
	}
	for (node = Head, i = 0; node; node = next) {
		next = node->next;
		if (S_ISREG(node->mode))
			list[i++] = node;
//		else
//			free_node(node);
//@@here, make only dir list, free later
	}

	/* sort by size, dev, ino, disp */
	if (debug) fprintf(stderr, "sort by size\n");
	qsort(list, n, sizeof(list[0]), comp_size_ino);
#define DEBUG_LIST()				\
 printf("\n");					\
 for (i = 0; i < n; i++)			\
  printf("size=%ld ino=%ld disp=%d name=%s\n",	\
   list[i]->size, list[i]->ino, list[i]->disp, list[i]->name);
//DEBUG_LIST();

	/* remove dup ino */
	/* TODO: ino-less device */
	if (debug) fprintf(stderr, "exclude duplicate path ");
	for (i = 0; i < n; i++)
		if (i + 1 < n) {
#ifdef _WIN32
			char path1[MAXPATH], path2[MAXPATH];

			setpath(path, list[i]->name, list[i]->parent);
			GetFullPathName(path, sizeof(path1), path1, NULL);
			setpath(path, list[i + 1]->name, list[i + 1]->parent);
			GetFullPathName(path, sizeof(path2), path2, NULL);
			if (stricmp(path1, path2) == 0)
#else
			if (list[i]->dev == list[i + 1]->dev
				&& list[i]->ino == list[i + 1]->ino)
#endif
			{
				free_node(list[i]);
				list[i] = NULL;
			}
		}
	for (i = 0, j = 0; i < n; i++)
		if (list[i] != NULL)
			list[j++] = list[i];
	n = j;
	if (debug) fprintf(stderr, "(%d)\n", n);
//DEBUG_LIST();

	/* remove nodisp files */
	if (debug) fprintf(stderr, "exclude no focus size ");
	for (i = 0; i < n; ) {
		int d;

		for (d = 0, j = i; j < n && list[i]->size == list[j]->size; j++)
			d |= list[j]->disp;
		if (!d)
			for (; i < j; i++) {
				free_node(list[i]);
				list[i] = NULL;
			}
		else
			i = j;
	}
	for (i = 0, j = 0; i < n; i++)
		if (list[i] != NULL)
			list[j++] = list[i];
	n = j;
	if (debug) fprintf(stderr, "(%d)\n", n);
//DEBUG_LIST();

	/* remove zero size */
	if (!all) {
		if (debug) fprintf(stderr, "exclude zero size ");
		for (i = 0; i < n; i++) {
			if (list[i]->size != 0)
				continue;
			free_node(list[i]);
			list[i] = NULL;
		}
		for (i = 0, j = 0; i < n; i++)
			if (list[i] != NULL)
				list[j++] = list[i];
		n = j;
		if (debug) fprintf(stderr, "(%d)\n", n);
	}

	/* remove unique size */
	if (!all && !uniq) {
		if (debug) fprintf(stderr, "exclude unique size ");
		for (i = 0; i < n; i++) {
			if (i + 1 < n && list[i]->size == list[i + 1]->size) {
				while (i + 1 < n && list[i]->size == list[i + 1]->size)
					i++;
				continue;
			}
			free_node(list[i]);
			list[i] = NULL;
		}
		for (i = 0, j = 0; i < n; i++)
			if (list[i] != NULL)
				list[j++] = list[i];
		n = j;
		if (debug) fprintf(stderr, "(%d)\n", n);
	}
	if (n == 0)
		goto end;

	/* compare and mark */
	if (debug) fprintf(stderr, "compare files\n");
	for (i = 0; i < n; i++) {
		list[i]->next = NULL;
		list[i]->prev = NULL;
#ifdef BLOCK_HASH
		j = (list[i]->size + BUFSIZE - 1) / BUFSIZE;
		if (j && (list[i]->block = malloc(sizeof(list[0]->block[0]) * j)) == NULL) {
			perror(Arg0);
			exit(1);
		}
#endif
	}
	qsort(list, n, sizeof(list[0]), comp_file);

	/* sort by disp, path */
	if (debug) fprintf(stderr, "sort by path\n");
	qsort(list, n, sizeof(list[0]), comp_disp_path);

	/* collect groups */
	if (debug) fprintf(stderr, "collect groups\n");
	for (i = 0; i < n; i++) {
		struct node *n, *n2;

		n = list[i];
		if (n->prev != NULL || (n->next != NULL && n->next->next != n->next)) {
			while (n->prev != NULL)
				n = n->prev;
			while (n != NULL) {
				n2 = n;
				n = n->next;
				n2->next = list[i];
				n2->prev = NULL;
			}
		}
	}

	/* display */
	if (debug) fprintf(stderr, "print files\n");
	for (i = 0; i < n; i++) {
		if (list[i]->next == NULL) {
			if ((all || uniq) && list[i]->disp) {
				if (len)
					printf("%7lu\t", list[i]->size);
				setpath(path, list[i]->name, list[i]->parent);
				printf("%s\n", path);
			}
			continue;
		}
		for (j = i; j < n; j++)
			if (list[j]->next == list[i] && list[j]->disp)
				break;
		if (j == n)
			continue;
		for (j = i; j < n; j++)
			if ((all || !uniq) && list[j]->next == list[i]) {
				if (i != j)
					printf("\t");
				else if (len)
					printf("%7lu\t", list[j]->size);
				setpath(path, list[j]->name, list[j]->parent);
				printf("%s\n", path);
			}
	}

end:
	if (debug) fprintf(stderr, "clean\n");
	for (i = 0; i < n; i++)
		free_node(list[i]);
	if (list != NULL)
		free(list);

	return 0;
}

